@@ -27,18 +27,18 @@
 #include "parse_keyword.h"
 #include "dispatch_key.h"
 #include <stdlib.h>
-#include <unistd.h>
-#include <sys/uio.h>
 #include <string.h>
 #include <stdio.h>
-#include <errno.h>
-#include <sys/time.h>
-#include <sys/types.h>
-#include <sys/socket.h>
+#ifndef WIN32
+#include "socket_posix.h"
+#include <sys/uio.h>
 #include <signal.h>
 #include <time.h>
 #include <netinet/in.h>
 #include <netinet/tcp.h>
+#else  /* WIN32 */
+#include "socket_win32.h"
+#endif  /* WIN32 */
 
 
 #ifndef MAX_IOVEC
@@ -104,6 +104,7 @@ struct command_state
 {
   struct client *client;
   int fd;
+  struct pollfd *pollfd;
   enum socket_mode_e socket_mode;
   int noreply;
   int last_cmd_noreply;
@@ -248,6 +249,7 @@ struct index_node
 
 struct client
 {
+  struct array pollfds;
   struct array servers;
 
   struct dispatch_state dispatch;
@@ -340,10 +342,18 @@ next_index(struct command_state *state)
 struct client *
 client_init()
 {
-  struct client *c = malloc(sizeof(struct client));
+  struct client *c;
+
+#ifdef WIN32
+  if (win32_socket_library_acquire() != 0)
+    return NULL;
+#endif  /* WIN32 */
+
+  c = malloc(sizeof(struct client));
   if (! c)
     return NULL;
 
+  array_init(&c->pollfds);
   array_init(&c->servers);
   array_init(&c->index_list);
   array_init(&c->str_buf);
@@ -388,12 +398,17 @@ client_destroy(struct client *c)
   dispatch_destroy(&c->dispatch);
 
   array_destroy(&c->servers);
+  array_destroy(&c->pollfds);
   array_destroy(&c->index_list);
   array_destroy(&c->str_buf);
 
   if (c->prefix_len > 1)
     free(c->prefix);
   free(c);
+
+#ifdef WIN32
+  win32_socket_library_release();
+#endif  /* WIN32 */
 }
 
 
@@ -413,14 +428,14 @@ client_set_ketama_points(struct client *c, int ketama_points)
 void
 client_set_connect_timeout(struct client *c, int to)
 {
-  c->connect_timeout = to;
+  c->connect_timeout = (to > 0 ? to : -1);
 }
 
 
 void
 client_set_io_timeout(struct client *c, int to)
 {
-  c->io_timeout = to;
+  c->io_timeout = (to > 0 ? to : -1);
 }
 
 
@@ -469,6 +484,9 @@ client_add_server(struct client *c, const char *host, size_t host_len,
   if (weight <= 0.0)
     return MEMCACHED_FAILURE;
 
+  if (array_extend(c->pollfds, struct pollfd, 1, ARRAY_EXTEND_EXACT) == -1)
+    return MEMCACHED_FAILURE;
+
   if (array_extend(c->servers, struct server, 1, ARRAY_EXTEND_EXACT) == -1)
     return MEMCACHED_FAILURE;
 
@@ -482,6 +500,7 @@ client_add_server(struct client *c, const char *host, size_t host_len,
   if (res == -1)
     return MEMCACHED_FAILURE;
 
+  array_push(c->pollfds);
   array_push(c->servers);
 
   return MEMCACHED_SUCCESS;
@@ -1370,11 +1389,9 @@ state_prepare(struct command_state *state)
 int
 client_execute(struct client *c)
 {
-  struct timeval to, *pto = c->io_timeout > 0 ? &to : NULL;
-  fd_set write_set, read_set;
   int first_iter = 1;
 
-#ifndef MSG_NOSIGNAL
+#if ! defined(MSG_NOSIGNAL) && ! defined(WIN32)
   struct sigaction orig, ignore;
   int res;
 
@@ -1384,14 +1401,17 @@ client_execute(struct client *c)
   res = sigaction(SIGPIPE, &ignore, &orig);
   if (res == -1)
     return MEMCACHED_FAILURE;
-#endif /* ! MSG_NOSIGNAL */
+#endif /* ! defined(MSG_NOSIGNAL) && ! defined(WIN32) */
 
   while (1)
     {
       struct server *s;
-      int max_fd, res;
+      struct pollfd *pollfd_beg, *pollfd;
+      int res;
+
+      pollfd_beg = array_beg(c->pollfds, struct pollfd);
+      pollfd = pollfd_beg;
 
-      max_fd = -1;
       for (array_each(c->servers, struct server, s))
         {
           int may_write, may_read;
@@ -1410,76 +1430,50 @@ client_execute(struct client *c)
             }
           else
             {
-              may_write = FD_ISSET(state->fd, &write_set);
-              may_read = FD_ISSET(state->fd, &read_set);
+              const short revents = state->pollfd->revents;
+
+              may_write = revents & (POLLOUT | POLLERR | POLLHUP);
+              may_read = revents & (POLLIN | POLLERR | POLLHUP);
             }
 
           if (may_read || may_write)
             {
-#if 1
-              /*
-                At least gcc 3.4.2--4.2.2 report that res may be used
-                uninitialized here.  Though this doesn't look so, we
-                initialize it to suppress the warning.
-              */
-              int res = 0;
-#endif
-
               if (may_write)
                 {
+                  int res;
+
                   res = send_request(state, s);
                   if (res == MEMCACHED_CLOSED)
                     may_read = 0;
                 }
 
               if (may_read)
-                res = process_reply(state, s);
+                process_reply(state, s);
 
-              if (is_active(state))
-                {
-                  if (max_fd < state->fd)
-                    max_fd = state->fd;
-                }
+              if (! is_active(state))
+                continue;
             }
-          else
+
+          pollfd->events = 0;
+
+          if (state->iov_count > 0)
+            pollfd->events |= POLLOUT;
+          if (state->reply_count > 0 || state->nowait_count > 0)
+            pollfd->events |= POLLIN;
+
+          if (pollfd->events != 0)
             {
-              if (max_fd < state->fd)
-                max_fd = state->fd;
+              pollfd->fd = state->fd;
+              state->pollfd = pollfd;
+              ++pollfd;
             }
         }
 
-      if (max_fd == -1)
+      if (pollfd == pollfd_beg)
         break;
 
-      FD_ZERO(&write_set);
-      FD_ZERO(&read_set);
-      for (array_each(c->servers, struct server, s))
-        {
-          struct command_state *state = &s->cmd_state;
-
-          if (is_active(state))
-            {
-              if (state->iov_count > 0)
-                FD_SET(state->fd, &write_set);
-              if (state->reply_count > 0 || state->nowait_count > 0)
-                FD_SET(state->fd, &read_set);
-            }
-        }
-
       do
-        {
-          /*
-            For maximum portability across systems that may or may not
-            modify the timeout argument we treat it as undefined after
-            the call, and reinitialize on every iteration.
-          */
-          if (pto)
-            {
-              pto->tv_sec = c->io_timeout / 1000;
-              pto->tv_usec = (c->io_timeout % 1000) * 1000;
-            }
-          res = select(max_fd + 1, &read_set, &write_set, NULL, pto);
-        }
+        res = poll(pollfd_beg, pollfd - pollfd_beg, c->io_timeout);
       while (res == -1 && errno == EINTR);
 
       /*
@@ -1511,13 +1505,13 @@ client_execute(struct client *c)
       first_iter = 0;
     }
 
-#ifndef MSG_NOSIGNAL
+#if ! defined(MSG_NOSIGNAL) && ! defined(WIN32)
   /*
     Ignore return value of sigaction(), there's nothing we can do in
     the case of error.
   */
   sigaction(SIGPIPE, &orig, NULL);
-#endif /* ! MSG_NOSIGNAL */
+#endif /* ! defined(MSG_NOSIGNAL) && ! defined(WIN32) */
 
   return MEMCACHED_SUCCESS;
 }
@@ -1538,7 +1532,7 @@ tcp_optimize_latency(struct command_state *state)
     {
       static const int enable = 1;
       setsockopt(state->fd, IPPROTO_TCP, TCP_NODELAY,
-                 &enable, sizeof(enable));
+                 (void *) &enable, sizeof(enable));
       state->socket_mode = TCP_LATENCY;
     }
 #endif /* TCP_NODELAY */
@@ -1554,7 +1548,7 @@ tcp_optimize_throughput(struct command_state *state)
     {
       static const int disable = 0;
       setsockopt(state->fd, IPPROTO_TCP, TCP_NODELAY,
-                 &disable, sizeof(disable));
+                 (void *) &disable, sizeof(disable));
       state->socket_mode = TCP_THROUGHPUT;
     }
 #endif /* TCP_NODELAY */
@@ -1585,7 +1579,7 @@ get_server_fd(struct client *c, struct server *s)
       if (s->port)
         {
           state->fd = client_connect_inet(s->host, s->port,
-                                          1, c->connect_timeout);
+                                          c->connect_timeout);
           /* This is to trigger actual reset.  */
           state->socket_mode = TCP_THROUGHPUT;
           tcp_optimize_latency(state);
